# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.components import Result
from suanpan.storage.arguments import Folder


class CommonModel(Folder):
    def __init__(self, key, type, version="latest", **kwargs):
        super(CommonModel, self).__init__(key, **kwargs)
        self.version = version
        self.modelType = type
        self.model = self.modelType()

    def load(self, args):
        super(CommonModel, self).load(args)
        if self.folderPath:
            self.value = self.model
        return self.value

    def format(self, context):
        super(CommonModel, self).format(context)
        if self.folderPath:
            self.model.load(self.folderPath)
            self.value = self.model
        return self.value

    def clean(self, context):
        super(CommonModel, self).clean(context)
        if self.folderPath:
            self.value = self.model
        return self.value

    def save(self, context, result):
        model = result.value
        model.save(self.folderPath)
        return super(CommonModel, self).save(
            context, Result.froms(value=self.folderPath)
        )


class HotReloadModel(CommonModel):
    def format(self, context):
        if self.folderName:
            self.model.loadFrom(self.folderName, version=self.version)
            self.value = self.model
        return self.value

    def save(self, context, result):
        raise NotImplementedError("Not support save!")



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import path
from suanpan.arguments import Arg
from suanpan.log import logger
from suanpan.state.storage import StorageIndexSaver
from suanpan.storage import storage


class StateArg(Arg):
    pass


class StorageLoaderArg(StateArg):
    STATE_CLASS = StorageIndexSaver
    STATE_PATTERN = "storage_{index}"

    def __init__(self, *args, **kwargs):
        super(StorageLoaderArg, self).__init__(*args, **kwargs)
        self.loader = None
        self.folderName = None
        self.folderPath = None

    def load(self, args):
        self.folderName = super(StorageLoaderArg, self).load(args)
        if self.folderName:
            self.folderPath = storage.getPathInTempStore(self.folderName)
            self.loader = self.STATE_CLASS(
                name=self.folderName, pattern=self.STATE_PATTERN
            )
        if self.folderPath:
            path.mkdirs(self.folderPath)
        self.value = self.loader
        return self.value

    def clean(self, context):
        raise NotImplementedError(
            "{} can't be set as output argument.".format(self.name)
        )


class StorageSaverArg(StateArg):
    STATE_CLASS = StorageIndexSaver
    STATE_PATTERN = "storage_{index}"

    def __init__(self, *args, **kwargs):
        self.saver = None
        self.folderName = None
        self.folderPath = None
        self.stateClass = kwargs.pop("type", self.STATE_CLASS)
        self.statePattern = kwargs.pop("pattern", self.STATE_PATTERN)
        super(StorageSaverArg, self).__init__(*args, **kwargs)

    def load(self, args):
        self.folderName = super(StorageSaverArg, self).load(args)
        if self.folderName:
            self.folderPath = storage.getPathInTempStore(self.folderName)
            self.saver = self.stateClass(
                name=self.folderName, pattern=self.statePattern
            )
        if self.folderPath:
            path.mkdirs(self.folderPath)
        self.value = self.saver
        return self.value

    def clean(self, context):
        if self.folderName:
            storage.remove(self.folderName)
        if self.folderPath:
            path.empty(self.folderPath)
        return self.folderPath

    def save(self, context, result):
        logger.info("Saving: ({}) Nothing to do!".format(self.name))



# coding=utf-8
from __future__ import absolute_import, print_function

import os

from suanpan.stream import Stream


class StreamTester(Stream):
    def __init__(self, *args, **kwargs):
        os.environ["STREAM_SEND_QUEUE"] = os.environ["STREAM_RECV_QUEUE"]
        super(StreamTester, self).__init__(*args, **kwargs)


tester = StreamTester()


def send(*args, **kwargs):
    return tester.sendMissionMessage(tester.generateMessage(), *args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import app
from suanpan import arguments as base  # pylint: disable=unused-import
from suanpan.arguments import auto  # pylint: disable=unused-import
from suanpan.dw import arguments as dw
from suanpan.imports import imports
from suanpan.storage import arguments as storage

Int = imports("suanpan.{}.arguments.Int".format(app.TYPE))
Float = imports("suanpan.{}.arguments.Float".format(app.TYPE))
Bool = imports("suanpan.{}.arguments.Bool".format(app.TYPE))
String = imports("suanpan.{}.arguments.String".format(app.TYPE))
List = imports("suanpan.{}.arguments.List".format(app.TYPE))
ListOfString = imports("suanpan.{}.arguments.ListOfString".format(app.TYPE))
ListOfInt = imports("suanpan.{}.arguments.ListOfInt".format(app.TYPE))
ListOfFloat = imports("suanpan.{}.arguments.ListOfFloat".format(app.TYPE))
ListOfBool = imports("suanpan.{}.arguments.ListOfBool".format(app.TYPE))
IntOrFloat = imports("suanpan.{}.arguments.IntOrFloat".format(app.TYPE))
IntFloatOrString = imports("suanpan.{}.arguments.IntFloatOrString".format(app.TYPE))
BoolOrString = imports("suanpan.{}.arguments.BoolOrString".format(app.TYPE))
StringOrListOfFloat = imports(
    "suanpan.{}.arguments.StringOrListOfFloat".format(app.TYPE)
)
Json = imports("suanpan.{}.arguments.Json".format(app.TYPE))

File = imports("suanpan.{}.arguments.File".format(app.TYPE))
Folder = imports("suanpan.{}.arguments.Folder".format(app.TYPE))
Data = imports("suanpan.{}.arguments.Data".format(app.TYPE))
Csv = imports("suanpan.{}.arguments.Csv".format(app.TYPE))
Npy = imports("suanpan.{}.arguments.Npy".format(app.TYPE))
Visual = imports("suanpan.{}.arguments.Visual".format(app.TYPE))
Model = imports("suanpan.{}.arguments.Model".format(app.TYPE))
H5Model = imports("suanpan.{}.arguments.H5Model".format(app.TYPE))
Checkpoint = imports("suanpan.{}.arguments.Checkpoint".format(app.TYPE))
JsonModel = imports("suanpan.{}.arguments.JsonModel".format(app.TYPE))
Image = imports("suanpan.{}.arguments.Image".format(app.TYPE))

Table = imports("suanpan.{}.arguments.Table".format(app.TYPE))
DataFrame = imports("suanpan.{}.arguments.DataFrame".format(app.TYPE))



import suanpan
from suanpan.app import app
from suanpan.app.arguments import Folder, String


@app.input(Folder(key="inputData1"))
@app.output(Folder(key="outputData1"))
@app.param(String(key="param1"))
def test(context):
    args = context.args
    print(args.inputData1)
    print(args.outputData1)
    print(args.param1)
    return args.inputData1


if __name__ == "__main__":
    suanpan.run(test)



# coding=utf-8
from __future__ import absolute_import, print_function

import suanpan
from suanpan.utils import json
from tests.test_auto_args import Demo

env = {"STORAGE_TYPE": "local", "STORAGE_LOCAL_TEMP_STORE": "tmp"}
params = {
    "inputData1": "input_test",
    "outputData1": "output_test",
    "param1": json.dumps({"b": 2}),
}

suanpan.run("tests.test_auto_args.Demo", env=env, **params)

suanpan.run(Demo, env=env, **params)

with suanpan.env(**env):
    suanpan.run(Demo, **params)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.state.base import IndexSaver, TimeSaver
from suanpan.state.storage import StorageSaver
from suanpan.storage import storage
from suanpan.utils import image


class ScreenshotsSaver(StorageSaver):
    def update(self, imageOrPath):
        data = image.read(imageOrPath) if isinstance(imageOrPath, str) else imageOrPath
        self.current.localPath = storage.getPathInTempStore(self.current.storageName)
        storageName = storage.storagePathJoin(
            self.current.storageName, self.currentPattern
        )
        localPath = storage.localPathJoin(self.current.localPath, self.currentPattern)
        image.save(localPath, data, self.current.get("flag"))
        return storage.upload(storageName, localPath)


class ScreenshotsIndexSaver(ScreenshotsSaver, IndexSaver):
    PATTERN = "screenshot_{index}.png"


class ScreenshotsTimeSaver(ScreenshotsSaver, TimeSaver):
    PATTERN = "screenshot_{time}.png"



# coding=utf-8
from __future__ import absolute_import, print_function

import base64
import hashlib
import os
import shutil

from suanpan import asyncio


def safeMkdirs(path):
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except FileExistsError:
            pass
    return path


def safeMkdirsForFile(filepath):
    return safeMkdirs(os.path.dirname(os.path.abspath(filepath)))


def mkdirs(path, parent=False):
    mkdirFunction = safeMkdirsForFile if parent else safeMkdirs
    return mkdirFunction(path)


def merge(paths, dist, mode="move"):
    safeMkdirs(dist)
    mergeFunc = getattr(shutil, mode)
    for path in paths:
        if os.path.isfile(path):
            mergeFunc(path, dist)
        else:
            for p in os.listdir(path):
                subpath = os.path.join(path, p)
                mergeFunc(subpath, dist)
            if mode == "move":
                os.rmdir(path)
    return dist


def md5(filepath, block_size=64 * 1024):
    with open(filepath, "rb") as f:
        _md5 = hashlib.md5()
        while True:
            data = f.read(block_size)
            if not data:
                break
            _md5.update(data)
    return base64.b64encode(_md5.digest()).decode()


def empty(folder):
    def _safeListDir(folder):
        try:
            return os.listdir(folder)
        except OSError:
            return []

    for path in _safeListDir(folder):
        remove(os.path.join(folder, path))
    return folder


def remove(path):
    removeFunction = removeFile if os.path.isfile(path) else removeFolder
    return removeFunction(path)


def removeFile(filePath):
    if os.path.exists(filePath):
        try:
            os.remove(filePath)
        except (FileNotFoundError, OSError):
            pass
    return filePath


def removeFolder(folderPath):
    if os.path.exists(folderPath):
        try:
            shutil.rmtree(folderPath)
        except (FileNotFoundError, OSError):
            pass
    return folderPath


def removeEmptyFolders(folderPath):
    try:
        os.removedirs(folderPath)
    except OSError:
        pass
    return folderPath


def copy(src, dst):
    if not os.path.exists(src):
        raise FileNotFoundError(src)
    copyFunction = copyFile if os.path.isfile(src) else copyFolder
    return copyFunction(src, dst)


def copyFile(src, dst):
    mkdirs(dst, parent=True)
    shutil.copyfile(src, dst)
    return dst


def copyFolder(src, dst):
    remove(dst)
    mkdirs(dst, parent=True)
    shutil.copytree(src, dst)
    return dst


def getPathDepth(root, path):
    root = root.rstrip(os.sep)
    path = path.rstrip(os.sep)
    return path[len(root) :].count(os.sep)


def listdir(path, depth=1):
    folders, files = [], []
    for root, folders, files in os.walk(path):
        if depth is True or getPathDepth(path, root) < depth:
            folders.extend([os.path.join(root, folder) for folder in folders])
            files.extend([os.path.join(root, file) for file in files])
    return folders, files


def replaceBasePath(src, dst, *paths):
    src = src.rstrip(os.sep)
    dst = dst.rstrip(os.sep)
    return [path.rstrip(os.sep).replace(src, dst) for path in paths]


# Concurrent Funcs


def cremove(path, workers=None):
    removeFunction = cremoveFile if os.path.isfile(path) else cremoveFolder
    return removeFunction(path, workers=workers)


def cremoveFile(filePath, workers=None):  # pylint: disable=unused-argument
    return removeFile(filePath)


def cremoveFolder(folderPath, workers=None):
    cempty(folderPath, workers=workers)
    return remove(folderPath)


def cempty(folderPath, workers=None):
    folders, files = listdir(folderPath, depth=True)
    asyncio.map(remove, files, pbar="Removing Files", workers=workers)
    asyncio.map(remove, folders, pbar="Removing Folders", workers=workers)
    return folderPath


def ccopy(src, dst, workers=None):
    if not os.path.exists(src):
        raise FileNotFoundError(src)
    copyFunction = ccopyFile if os.path.isfile(src) else ccopyFolder
    return copyFunction(src, dst, workers=workers)


def ccopyFile(src, dst, workers=None):  # pylint: disable=unused-argument
    return copyFile(src, dst)


def ccopyFolder(src, dst, workers=None):
    folders, files = listdir(src, depth=True)
    newfolders = replaceBasePath(src, dst, *folders)
    newfiles = replaceBasePath(src, dst, *files)
    mkdirs(dst)
    asyncio.map(mkdirs, newfolders, pbar="Making Folders", workers=workers)
    asyncio.starmap(copy, zip(files, newfiles), pbar="Copying Files", workers=workers)
    return dst
